import typing

import numpy
import torch
import sklearn.base
import sklearn.metrics

import cvtda.utils
import cvtda.logging
import cvtda.neural_network


class MultiImageDataset:
    def __init__(self, features, labels, images):
        self.labels = torch.tensor(labels, dtype = torch.long)
        self.features = torch.tensor(features, dtype = torch.float32)
        self.images = [ torch.tensor(imgs, dtype = torch.float32).permute((0, 3, 1, 2)) for imgs in images ]


class MultiImageClassifier(sklearn.base.ClassifierMixin):
    def __init__(
        self,

        n_jobs: int = -1,
        random_state: int = 42,

        device: torch.device = torch.device("cuda"),
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        n_epochs: int = 100,
        
        skip_images: bool = False,
        skip_features: bool = False
    ):
        self.n_jobs_ = n_jobs
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs

        self.skip_images_ = skip_images
        self.skip_features_ = skip_features


    def fit(self, train: MultiImageDataset, val: typing.Optional[MultiImageDataset]):
        cvtda.utils.set_random_seed(self.random_state_)
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.labels, train.features, *train.images),
            batch_size = self.batch_size_,
            shuffle = True
        )
        self.init_(*next(iter(train_dl)))

        for epoch in range(self.n_epochs_):
            sum_loss = 0

            self.model_list_.train()
            for (labels, features, *images) in cvtda.logging.logger().pbar(train_dl, desc = f"Epoch {epoch}"):
                self.optimizer_.zero_grad()
                pred = self.forward_(features, *images)
                loss = torch.nn.functional.cross_entropy(pred, labels.to(self.device_), reduction = 'mean')
                loss.backward()
                self.optimizer_.step()
                sum_loss += loss.item()
            postfix = { 'loss': sum_loss, 'lr': self.optimizer_.param_groups[0]['lr'] }
            self.scheduler_.step()

            if val is not None:
                val_proba = self.predict_proba_(val)
                val_pred = numpy.argmax(val_proba, axis = 1)
                postfix['val_acc'] = sklearn.metrics.accuracy_score(val.labels, val_pred)

            cvtda.logging.logger().print(f"Epoch {epoch}:", postfix)

        self.fitted_ = True
        return self
    
    def predict_proba(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        assert self.fitted_ is True, 'fit() must be called before predict_proba()'
        cvtda.utils.set_random_seed(self.random_state_)
        return self.predict_proba_(dataset)


    def init_(
        self, 
        labels: torch.Tensor,
        features: torch.Tensor,
        *images: typing.List[torch.Tensor]
    ):
        self.features_base_ = cvtda.neural_network.NNBase(
            num_diagrams = 0, skip_diagrams = True, skip_images = True, skip_features = False
        ).to(self.device_).train()

        self.images_bases_ = torch.nn.ModuleList([
            cvtda.neural_network.NNBase(
                num_diagrams = 0, skip_diagrams = True, skip_images = False, skip_features = True
            ).to(self.device_).train()
            for _ in range(len(images))
        ])

        self.model_ = torch.nn.Sequential(
            torch.nn.Dropout(0.4), torch.nn.LazyLinear(256), torch.nn.BatchNorm1d(256), torch.nn.GELU(),
            torch.nn.Dropout(0.3), torch.nn.Linear(256, 128), torch.nn.BatchNorm1d(128), torch.nn.GELU(),
            torch.nn.Dropout(0.2), torch.nn.Linear(128, 64), torch.nn.BatchNorm1d(64), torch.nn.GELU(),
            torch.nn.Dropout(0.1), torch.nn.Linear(64, 32), torch.nn.BatchNorm1d(32), torch.nn.GELU(),
            torch.nn.Linear(32, len(torch.unique(labels)))
        ).to(self.device_).train()

        self.model_list_ = torch.nn.ModuleList([ self.features_base_, self.images_bases_, self.model_ ])

        self.optimizer_ = torch.optim.AdamW(
            params = self.model_list_.parameters(),
            lr = self.learning_rate_ * 100
        )
        
        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001
        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        features_result = self.features_base_(None, features.to(self.device_))
        images_result = [ self.images_bases_[i](images[i].to(self.device_), None) for i in range(len(images)) ]

        result = []
        if not self.skip_features_:
            result.append(features_result)
        if not self.skip_images_:
            result.extend(images_result)
        self.model_(torch.cat(result, dim = 1))

        cvtda.logging.logger().print(f'Input to LazyLinear: {self.model_[1].in_features}')
        cvtda.logging.logger().print(f'Parameters: {sum(p.numel() for p in self.model_list_.parameters())}')

    def forward_(
        self,
        features: torch.Tensor,
        *images: typing.List[torch.Tensor]
    ) -> torch.Tensor:
        result = []

        if not self.skip_features_:
            result.append(self.features_base_(None, features.to(self.device_)))

        if not self.skip_images_:
            result.extend([ self.images_bases_[i](images[i].to(self.device_), None) for i in range(len(images)) ])

        return self.model_(torch.cat(result, dim = 1))

    def predict_proba_(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.features, *dataset.images),
            batch_size = self.batch_size_,
            shuffle = False
        )

        y_pred_proba = [ ]
        self.model_list_.eval()
        with torch.no_grad():
            for (features, *images) in dl:
                y_pred_proba.append(self.forward_(features, *images))
        y_pred_proba = torch.vstack(y_pred_proba)
        return torch.nn.functional.softmax(y_pred_proba, dim = 1).cpu().numpy()



import torch

class Dataset:
    def __init__(self, images, features, masks):
        self.images = torch.tensor(images, dtype = torch.float32).permute((0, 3, 1, 2))
        self.masks = torch.tensor(masks, dtype = torch.float32).unsqueeze(1)
        self.features = torch.tensor(features, dtype = torch.float32)



import typing

import numpy
import torch
import torch.utils.data

import cvtda.utils
import cvtda.logging
import cvtda.neural_network

from .estimate_quality import estimate_quality

class Autoencoder:
    def __init__(
        self,

        random_state: int = 42,

        device: torch.device = torch.device("cuda"),
        batch_size: int = 256,
        learning_rate: float = 1e-3,
        n_epochs: int = 100,
        latent_dim: int = 256,
        
        skip_diagrams: bool = False,
        skip_images: bool = False,
        skip_features: bool = False
    ):
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs
        self.latent_dim_ = latent_dim

        self.skip_diagrams_ = skip_diagrams
        self.skip_images_ = skip_images
        self.skip_features_ = skip_features


    def fit(self, train: cvtda.neural_network.Dataset, val: typing.Optional[cvtda.neural_network.Dataset]):
        cvtda.utils.set_random_seed(self.random_state_)
        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, torch.arange(len(train))),
            batch_size = self.batch_size_,
            shuffle = True
        )
        self.init_(*next(iter(train_dl)), train)

        for epoch in range(self.n_epochs_):
            sum_loss = 0

            self.model_list_.train()
            for images, features, diagram_idxs in cvtda.logging.logger().pbar(train_dl, desc = f"Epoch {epoch}"):
                self.optimizer_.zero_grad()
                pred = self.decode_(self.encode_(images, features, diagram_idxs, train))
                loss = torch.nn.functional.mse_loss(pred, images.to(self.device_), reduction = 'mean')
                loss.backward()
                self.optimizer_.step()
                sum_loss += loss.item()
            postfix = { 'loss': sum_loss, 'lr': self.optimizer_.param_groups[0]['lr'] }
            self.scheduler_.step()

            if val is not None:
                decoded = self.decode(self.encode(val))
                postfix = { **postfix, **estimate_quality(decoded, val.images.permute((0, 2, 3, 1)).cpu().numpy()) }

            cvtda.logging.logger().print(f"Epoch {epoch}:", postfix)

        self.fitted_ = True
        return self
    
    def encode(self, dataset: cvtda.neural_network.Dataset) -> numpy.ndarray:
        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.images, dataset.features, torch.arange(len(dataset))),
            batch_size = self.batch_size_,
            shuffle = False
        )

        encoded = [ ]
        self.model_list_.eval()
        with torch.no_grad():
            for images, features, diagram_idxs in dl:
                encoded.append(self.encode_(images, features, diagram_idxs, dataset))
        return torch.vstack(encoded).cpu().numpy()

    
    def decode(self, latent_vectors: numpy.ndarray) -> numpy.ndarray:
        dl = torch.utils.data.DataLoader(latent_vectors, batch_size = self.batch_size_, shuffle = False)

        decoded = [ ]
        self.model_list_.eval()
        with torch.no_grad():
            for vecs in dl:
                decoded.append(self.decode_(vecs))
        return torch.vstack(decoded).permute((0, 2, 3, 1)).squeeze().cpu().numpy()


    def init_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ):
        diagrams = [] if self.skip_diagrams_ else dataset.get_diagrams(diagram_idxs)
        self.encoder_base_ = cvtda.neural_network.NNBase(
            num_diagrams = len(diagrams) // 2,
            skip_diagrams = self.skip_diagrams_,
            skip_images = self.skip_images_,
            skip_features = self.skip_features_,
            images_n_channels = images.shape[1]
        ).to(self.device_).train()
        self.encoder_ = torch.nn.LazyLinear(self.latent_dim_).to(self.device_).train()

        w = images.shape[2] // 4
        h = images.shape[3] // 4
        self.decoder_ = torch.nn.Sequential(
            torch.nn.InstanceNorm1d(self.latent_dim_), torch.nn.Linear(self.latent_dim_, 128 * w * h),
            torch.nn.InstanceNorm1d(128 * w * h), torch.nn.Unflatten(1, (128, w, h)), torch.nn.ReLU(),
            torch.nn.ConvTranspose2d(128, 64, kernel_size = 3, stride = 2, output_padding = 1, padding = 1), torch.nn.InstanceNorm2d(64), torch.nn.ReLU(),
            torch.nn.Conv2d(64, 64, kernel_size = 3, padding = 1), torch.nn.InstanceNorm2d(64), torch.nn.ReLU(),
            torch.nn.ConvTranspose2d(64, 32, kernel_size = 3, stride = 2, output_padding = 1, padding = 1), torch.nn.InstanceNorm2d(32), torch.nn.ReLU(),
            torch.nn.Conv2d(32, 16, kernel_size = 3, padding = 1), torch.nn.InstanceNorm2d(16), torch.nn.ReLU(),
            torch.nn.Conv2d(16, images.shape[1], kernel_size = 3, padding = 1), torch.nn.Sigmoid()
        ).to(self.device_).train()

        self.model_list_ = torch.nn.ModuleList([ self.encoder_base_, self.encoder_, self.decoder_ ])

        self.optimizer_ = torch.optim.AdamW(
            params = self.model_list_.parameters(),
            lr = self.learning_rate_ * 100
        )
        
        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001
        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        encoded = self.encode_(images, features, diagram_idxs, dataset)
        assert encoded.shape == (images.shape[0], self.latent_dim_), encoded.shape

        decoded = self.decode(encoded)
        assert decoded.shape == images.permute((0, 2, 3, 1)).squeeze().shape, f'{decoded.shape} != {images.shape}'

        cvtda.logging.logger().print(f'Encoder Base parameters: {sum(p.numel() for p in self.encoder_base_.parameters())}')
        cvtda.logging.logger().print(f'Decoder parameters: {sum(p.numel() for p in self.decoder_.parameters())}')

    def encode_(
        self,
        images: torch.Tensor,
        features: torch.Tensor,
        diagram_idxs: torch.Tensor,
        dataset: cvtda.neural_network.Dataset
    ) -> torch.Tensor:
        images = images.to(self.device_)
        features = features.to(self.device_)
        if self.skip_diagrams_:
            return self.encoder_(self.encoder_base_(images, features))
        return self.encoder_(self.encoder_base_(images, features, *dataset.get_diagrams(diagram_idxs)))

    def decode_(self, latent_vector: torch.Tensor) -> torch.Tensor:
        latent_vector = latent_vector.to(self.device_)
        return self.decoder_(latent_vector)



import typing

import numpy
import torch
import torch.utils.data
import segmentation_models_pytorch

import cvtda.utils
import cvtda.logging

from .impl import MiniUnetModule
from .estimate_quality import estimate_quality
from .Dataset import Dataset


class MiniUnet:
    def __init__(
        self,

        random_state: int = 42,

        device: torch.device = torch.device("cuda"),
        batch_size: int = 64,
        learning_rate: float = 1e-4,
        n_epochs: int = 100,
        remove_cross_maps: bool = False,
        
        with_images: bool = False,
        with_features: bool = False
    ):
        self.random_state_ = random_state

        self.device_ = device
        self.batch_size_ = batch_size
        self.learning_rate_ = learning_rate
        self.n_epochs_ = n_epochs
        self.remove_cross_maps_ = remove_cross_maps

        self.with_images_ = with_images
        self.with_features_ = with_features


    def fit(self, train: Dataset, val: typing.Optional[Dataset]):
        cvtda.utils.set_random_seed(self.random_state_)

        train_dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(train.images, train.features, train.masks),
            batch_size = self.batch_size_,
            shuffle = True
        )
        self.init_(*next(iter(train_dl)))

        pbar = cvtda.logging.logger().pbar(range(self.n_epochs_))
        for epoch in pbar:
            sum_loss = 0

            self.model_.train()
            for images, features, masks in train_dl:
                self.optimizer_.zero_grad()
                pred = self.forward_(images, features)
                loss = self.loss_(pred, masks.to(self.device_))
                loss.backward()
                self.optimizer_.step()
                sum_loss += loss.item()
            postfix = { 'loss': sum_loss, 'lr': self.optimizer_.param_groups[0]['lr'] }
            self.scheduler_.step()

            if val is not None:
                val_proba = self.predict_proba(val)
                postfix = { **postfix, **estimate_quality(val_proba, val.masks.cpu().numpy()) }
            cvtda.logging.logger().set_pbar_postfix(pbar, postfix)

        self.fitted_ = True
        return self
    
    def predict_proba(self, dataset: Dataset) -> numpy.ndarray:
        dl = torch.utils.data.DataLoader(
            torch.utils.data.TensorDataset(dataset.images, dataset.features),
            batch_size = self.batch_size_,
            shuffle = False
        )

        y_pred = [ ]
        self.model_.eval()
        with torch.no_grad():
            for images, features in dl:
                y_pred.append(self.forward_(images, features))
        return torch.vstack(y_pred).cpu().numpy()


    def init_(self, images: torch.Tensor, features: torch.Tensor, masks: torch.Tensor):
        self.model_ = MiniUnetModule(
            images,
            features,
            with_images = self.with_images_,
            with_features = self.with_features_,
            remove_cross_maps = self.remove_cross_maps_
        ).to(self.device_)

        self.optimizer_ = torch.optim.AdamW(params = self.model_.parameters(), lr = self.learning_rate_ * 100)
        
        def lr_scheduler_lambda(epoch):
            if epoch < self.n_epochs_ // 10:
                return 1
            if epoch < self.n_epochs_ // 4:
                return 0.1
            if epoch < self.n_epochs_ // 2:
                return 0.01
            if epoch < 3 * self.n_epochs_ // 4:
                return 0.001
            return 0.0001
        self.scheduler_ = torch.optim.lr_scheduler.LambdaLR(self.optimizer_, lr_scheduler_lambda)

        assert self.forward_(images, features).shape == masks.shape
        cvtda.logging.logger().print(f'Parameters: {sum(p.numel() for p in self.model_.parameters())}')

    def forward_(self, images: torch.Tensor, features: torch.Tensor) -> torch.Tensor:
        return self.model_(images.to(self.device_), features.to(self.device_))

    def loss_(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        bce = torch.nn.functional.binary_cross_entropy(input, target, reduction = 'mean')
        iou = segmentation_models_pytorch.losses.JaccardLoss('binary', from_logits = False)(input, target)
        return bce + iou



import typing

import numpy
import torch
import pandas

import cvtda.logging
import cvtda.dumping

from .Dataset import Dataset
from .MiniUnet import MiniUnet
from .estimate_quality import estimate_quality

def segment(
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_masks: numpy.ndarray,

    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_masks: numpy.ndarray,

    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,

    device: torch.device = torch.device('cuda'),
    batch_size: int = 64,
    learning_rate: float = 1e-3,
    n_epochs: int = 100,
    remove_cross_maps: bool = False
):
    nn_train = Dataset(train_images, train_features, train_masks)
    nn_test = Dataset(test_images, test_features, test_masks)

    def try_one(model: MiniUnet, name: str, display_name: str):
        cvtda.logging.logger().print(f'Trying {name} - {model}')
        
        dumper = cvtda.dumping.dumper()
        model_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
        if only_get_from_dump or dumper.has_dump(model_dump_name):
            y_pred_proba = dumper.get_dump(model_dump_name)
        else:
            model.fit(nn_train, nn_test)
            y_pred_proba = model.predict_proba(nn_test)

            if model_dump_name is not None:
                dumper.save_dump(y_pred_proba, model_dump_name)

        result = { 'model': display_name, **estimate_quality(y_pred_proba, test_masks) }
        cvtda.logging.logger().print(result)
        return result

    unet_kwargs = dict(
        random_state = random_state,
        device = device,
        batch_size = batch_size,
        learning_rate = learning_rate,
        n_epochs = n_epochs,
        remove_cross_maps = remove_cross_maps
    )
    models = [
        MiniUnet(**unet_kwargs, with_images = False, with_features = False),
        MiniUnet(**unet_kwargs, with_images = True, with_features = False),
        MiniUnet(**unet_kwargs, with_images = False, with_features = True),
        MiniUnet(**unet_kwargs, with_images = True, with_features = True)
    ]

    names = [
        'no',
        'images',
        'topological',
        'combined'
    ]
    display_names = [
        'Без признаков',
        'Без топологических признаков',
        'Только топологические признаки',
        'Комбинированные признаки'
    ]

    return pandas.DataFrame([ try_one(*args) for args in zip(models, names, display_names) ])



import torch

from .conv import make_conv

class Bottom(torch.nn.Module):
    def __init__(
        self,
        images_example: torch.Tensor,
        features_example: torch.Tensor,
        with_images: bool,
        with_features: bool,
        in_channels: int,
        out_channels: int
    ):
        super().__init__()
        self.with_images = with_images
        self.with_features = with_features
        self.in_shape = images_example.shape[1:]

        if with_images and with_features:
            in_channels *= 2
        self.conv = make_conv(in_channels, out_channels)

        flat_shape = torch.nn.Flatten()(images_example).shape[1]
        self.features = torch.nn.Sequential(
            torch.nn.Linear(features_example.shape[1], 4096), torch.nn.BatchNorm1d(4096), torch.nn.GELU(),
            torch.nn.Linear(4096, 2048), torch.nn.BatchNorm1d(2048), torch.nn.GELU(),
            torch.nn.Linear(2048, 1024), torch.nn.BatchNorm1d(1024), torch.nn.GELU(),
            torch.nn.Linear(1024, flat_shape), torch.nn.BatchNorm1d(flat_shape), torch.nn.GELU()
        )
        
    def forward(self, images: torch.Tensor, features: torch.Tensor):
        if not self.with_features and not self.with_images:
            return self.conv(torch.zeros_like(images))

        if self.with_features:
            target_shape = (features.shape[0], *self.in_shape)
            features = self.features(features).reshape(*target_shape)

        data_in = []
        if self.with_images:
            data_in.append(images)
        if self.with_features:
            data_in.append(features)
        return self.conv(torch.cat(data_in, dim = 1))



import torch

from .conv import make_conv

class UpsampleBlock(torch.nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.up_conv = torch.nn.ConvTranspose2d(in_channels, out_channels, kernel_size = 2, stride = 2)
        self.conv = make_conv(2 * out_channels, out_channels)
    
    def forward(self, input: torch.Tensor, features: torch.Tensor) -> torch.Tensor:
        features = torch.cat([ features, self.up_conv(input) ], dim = 1)
        return self.conv(features)
